import ast
import re
from abc import abstractmethod

import os
from bs4 import BeautifulSoup, NavigableString
from pyparsing import *
from jinja2 import Environment, FileSystemLoader, TemplateNotFound
from termcolor import colored


class Value(object):
    def __init__(self, val):
        self.val = val

    def parsed(self):
        return self.expr

    @abstractmethod
    def expr(self):
        pass

    @abstractmethod
    def __str__(self):
        pass


class ExpressionValue(Value):
    def parsed(self):
        return ast.literal_eval(self.val)

    def expr(self):
        return self.val

    def __str__(self) -> str:
        return '{{ ' + self.expr() + ' }}'


class StringValue(Value):
    def expr(self):
        return '"{}"'.format(self.val.replace('"', "'"))

    def __str__(self) -> str:
        return self.val


class TranslatableStringValue(StringValue):
    def expr(self):
        return super().expr() + '|_'

    def __str__(self) -> str:
        return '{{ ' + self.expr() + ' }}'


loaded_themes = {}


def render_template(theme_folder, template_name, context):
    context = context or {}

    if theme_folder not in loaded_themes:
        loader = FileSystemLoader(theme_folder)

        env = Environment(loader=loader,
                          variable_start_string='<{',
                          variable_end_string='}>',
                          block_start_string='<%',
                          block_end_string='%>'
                          )

        def expr(val):
            return val.expr()

        env.filters['expr'] = expr

        loaded_themes[theme_folder] = env
    else:
        env = loaded_themes[theme_folder]

    template = env.get_template(template_name)

    return template.render(**context)


def render_tag(el, theme_name='default', replace=True):
    self_content = ""
    for child in el.children:
        if isinstance(child, NavigableString):
            self_content += "\n" + str(child)  # str(parse_val(str(child).strip()))
        else:
            print(type(child), child)
            self_content += "\n" + render_tag(child, replace=False)

    context = {'_': self_content}
    for key, val in el.attrs.items():
        context[key] = parse_val(val)

    m = re.match('^block:([a-z0-9_]+)$', el.name)
    if m:
        tag_template = '{{% block {} %}}{}{{% endblock %}}'.format(m.group(1), self_content)
    else:
        theme_folder = 'col/themes/{}'.format(theme_name)
        template_name = '{}.html'.format(el.name.replace('.', '/'))
        try:
            tag_template = render_template(theme_folder, template_name, context=context)
        except TemplateNotFound:
            print(colored(
                'Warning!', on_color='on_red'),
                'Can not load template: ',
                colored(template_name, on_color='on_blue'),
                'Theme folder:', theme_folder
            )
            tag_template = self_content

        if replace:
            el.replaceWith(tag_template)

    return tag_template


def parse_val(val):
    if len(val) and val[0] == '!':
        new_val = ast.literal_eval(val[1:])
    elif len(val) and val[0] == '$':
        new_val = TranslatableStringValue(val[1:])
    elif len(val) and val[0] == '@':
        new_val = ExpressionValue(val[1:])
    else:
        new_val = StringValue(val)
    return new_val


def unescape_html(html):
    html = html.replace('&lt;', '<')
    html = html.replace('&gt;', '>')
    html = html.replace('&amp;', '&')
    return html


def render_blocks(file_source):
    parts = file_source.split('<genius:blocks')
    if len(parts) > 0:
        content = []
        for idx, part in enumerate(parts):
            if idx == 0:
                content.append(part)
            else:
                sub = part.split('</genius:blocks>')
                if len(sub) != 2:
                    print('WARN: Error parsing template file: not closed <genius:blocks> tag?')
                    continue

                block_html = '<genius:blocks' + sub[0] + '</genius:blocks>'
                block = BeautifulSoup(block_html, "html.parser")
                theme = block.attrs.get('theme', 'default')

                for block_tag in block.children:
                    for subtag in block_tag.children:
                        if isinstance(subtag, str):
                            continue
                        render_tag(subtag, theme_name=theme)

                first_block = list(block.children)[0]
                first_block.unwrap()

                source = unescape_html(str(block))

                content.append(source)
                content.append(sub[1])

        file_source = ''.join(content)

    return file_source



import sys
import zipfile
from io import BytesIO
from json.decoder import JSONDecodeError

import requests


class ApiError(Exception):
    def __init__(self, response, *args) -> None:
        super().__init__(*args)
        self.response = response


class GeniusClient(object):
    def __init__(self, api_url, token) -> None:
        super().__init__()

        self.api_url = api_url
        self.token = token

    def _get(self, service, **kwargs):
        headers = {'Authorization': f'Token {self.token}'}
        response = requests.get(f'{self.api_url}{service}', headers=headers, **kwargs)
        self.handle_error(service, response)
        return response.json()

    def _post(self, service, **kwargs):
        headers = {'Authorization': f'Token {self.token}'}
        url = f'{self.api_url}{service}'
        response = requests.post(url, headers=headers, **kwargs)
        self.handle_error(service, response)

        return response

    def handle_error(self, service, response):
        if response.status_code >= 400:
            try:
                error = response.json()

                print('')

                if 'detail' in error:
                    print(error['detail'])
                else:
                    print(error)

            except JSONDecodeError:
                print(
                    f'Http error {response.status_code}: Unexpected error happened on Genius server when working with /{service}.')
                print(response.content)

            raise ApiError(response)

    def generate(self, files, collections=None):
        response = self._post('generate', data={
            'collections': ','.join(collections),
        }, files={
            'gen': files
        })

        return response.content



import atexit
import os
import signal
import sys
from glob import glob

import click
from termcolor import colored

from genius_cli.client import GeniusClient, ApiError
from genius_cli.utils import collect_files, extract_files, collect_app_names, migrate_db, install_deps, remove_db, \
    wait_for_file_changes, run_django, run_webpack, npm_install, get_watch_paths


def run():
    api_token = os.environ.get('GENIUS_TOKEN', None)
    features_env = os.environ.get('GENIUS_FEATURES', None)

    # if not api_token:
    #     print('No genius api token. Add GENIUS_TOKEN variable to your profile.')
    #     sys.exit(1)

    genius = GeniusClient(
        api_url=os.environ.get('ZMEI_URL', 'http://ng.genius-project.io:9000/api/'),
        token=api_token,
    )

    @click.group()
    @click.option('--src', default='.', help='Sources path')
    @click.option('--dst', default='.', help='Target path')
    def cli(**args):
        pass

    @cli.command(help='Deploy to hyper.sh')
    def config(**args):
        print('Deploy!')
        pass

    @cli.command(help='Deploy to hyper.sh')
    def deploy(**args):
        print('Deploy!')

    @cli.command(help='Generate and start app')
    def up(**args):
        print('Up!')
        gen(up=True, **args)

    @cli.command(help='Run application')
    @click.option('--nodejs', is_flag=True, help='Initialize nodejs dependencies')
    @click.option('--watch', is_flag=True, help='Watch for changes')
    @click.option('--webpack', is_flag=True, help='Run webpack with reload when generation ends')
    @click.option('--port', default='8000', help='Django host:port to run on')
    @click.option('--host', default=None, help='Django host:port to run on')
    def run(**kwargs):
        print('Run!')
        gen(run=True, **kwargs)

    @cli.command(help='Just generate the code')
    @click.argument('app', nargs=-1)
    def generate(app=None, **args):
        print('generate!')
        gen(install=True, app=app or [])

    @cli.command(help='Install project dependencies')
    def install():
        print('Install!')
        gen(install=True)

    @cli.group(help='Database related commands')
    def db(**args):
        pass

    @db.command(help='Creates database migrations and apply them')
    @click.argument('app', nargs=-1)
    def migrate(app, **args):
        print('Db migrate!', app)
        gen(auto=True, app=app)

    @db.command(help='db remove + db migrate')
    @click.argument('app', nargs=-1)
    def rebuild(app, **args):
        print('Db rebuild!', app)
        gen(rebuild=True, app=app)

    @db.command(help='Rollback all the migrations')
    @click.argument('app', nargs=-1)
    def remove(app, **args):
        print('Db remove!', app)

        gen(remove=True, app=app)

    def gen(auto=False,
            src='.',
            dst='.',
            install=False,
            rebuild=False,
            remove=False,
            app=None,
            run=False,
            webpack=False,
            nodejs=False,
            host=None,
            port=8000,
            watch=False,
            up=False
            ):

        if rebuild:
            auto = True
            remove = True

        if not host:
            host = '127.0.0.1:{}'.format(port)

        if up:
            install = True
            auto = True
            watch = True
            run = True

        if not app or len(app) == 0:
            app = collect_app_names()

        src = os.path.realpath(src)
        dst = os.path.realpath(dst)

        django_process = None
        webpack_process = None

        def emergency_stop():
            if django_process:
                os.killpg(os.getpgid(django_process.pid), signal.SIGTERM)
            if webpack_process:
                os.killpg(os.getpgid(webpack_process.pid), signal.SIGTERM)

        atexit.register(emergency_stop)

        for i in wait_for_file_changes(get_watch_paths(), watch=watch):
            print('--------------------------------------------')
            print('Generating ...')
            print('--------------------------------------------')

            files = collect_files(src)

            try:
                files = genius.generate(files, collections=app)

                extract_files(dst, files)

                if up and os.path.exists('react'):
                    webpack = True
                    nodejs = True

                if nodejs and os.path.exists('react'):
                    npm_install()

                if remove:
                    remove_db(apps=app)

                if install or rebuild:
                    django_process = install_deps(django_process)

                if auto:
                    migrate_db(apps=app)

                if run and not django_process:
                    django_process = run_django(run_host=host)

                if webpack and not webpack_process:
                    webpack_process = run_webpack()

            except ApiError:
                pass

            if watch:
                print(colored('> ', 'white', 'on_blue'), 'Watching for changes...')

    cli()



import hashlib
import io
import json
import os
import re
import signal
import sys
import zipfile
from glob import glob
from contextlib import contextmanager
import subprocess
from io import BytesIO
from shutil import rmtree
from tarfile import TarFile

from termcolor import colored
from time import sleep

from genius_cli.blocks import render_blocks


@contextmanager
def chdir(path):
    cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(cwd)


PACKAGE_JSON_TOKEN = '__generated__'


def write_generated_file(path, source):
    # special hack for settings.py->SECRET_KEY
    if path.endswith('app/settings.py'):
        if os.path.exists(path):
            with open(path, 'r') as f:
                found = re.search('^SECRET_KEY\s*=.*$', f.read(), flags=re.MULTILINE)
                if found:
                    current_key = found.group(0)
                    new_key = re.search('^\s*SECRET_KEY\s*=.*$', source, flags=re.MULTILINE).group(0)
                    source = source.replace(new_key, current_key)

    dirname = os.path.dirname(path)
    if dirname and len(dirname) and not os.path.exists(dirname):
        os.makedirs(dirname)

    chksum = source_hash(source)

    tag = f"generated: {chksum}"

    source_prefix = ''

    if path.endswith('.py') or path.endswith('requirements.txt') or path.endswith('Dockerfile'):
        source_prefix = f"# {tag}\n\n"

    elif path.endswith('.js') or path.endswith('.jsx'):
        source_prefix = f"//# {tag}\n\n"

    elif path.endswith('package.json'):
        chksum = source_hash(json.dumps(json.loads(source), indent=4))
        source = json.dumps({**json.loads(source), **{PACKAGE_JSON_TOKEN: chksum}}, indent=4)

    elif ('/templates/' in path or '/themes/' in path) and path.endswith('.html'):
        source_prefix = f"{{# {tag} #}}\n\n"

    if os.path.exists(path):
        generated, changed, file_checksum = is_generated_file(path)

        if generated:
            if changed:
                print(colored(' ! ', 'red', 'on_yellow'), ' ', path)
                return  # checksum failed. Changed by hands?
            else:
                if file_checksum == chksum:
                    return  # already up to date
        else:
            # print(f'skip: {path}')
            return  # file intentionally overridden. Do not overwrite!

    if not os.path.exists(path):
        print(colored(' + ', 'blue', 'on_green'), ' ', path)
    else:
        print(colored(' ~ ', 'blue', 'on_cyan'), ' ', path)
    with open(path, 'w') as f:
        if source_prefix:
            f.write(source_prefix)
        f.write(source)
        # print(f'write: {path}')


def is_generated_file(path):
    """
    Checks if file is generated, and is it changed
    :param path:

    :return: generated, changed, checksum
    """
    with open(path, 'r') as f:
        content = f.read()

    if path.endswith('package.json'):
        try:
            data = json.loads(content)
        except Exception as e:
            print('Can not read package.json. That\'s shouldn\'t happen as file is auto-generated.', e)
            data = {}

        try:
            file_chk_expected = data.get(PACKAGE_JSON_TOKEN)
        except KeyError:
            return False, False, None

        data = json.loads(content)
        if PACKAGE_JSON_TOKEN in data:
            del data[PACKAGE_JSON_TOKEN]
        real_source = json.dumps(data, indent=4)

    else:
        match = re.match('^({|//)?# generated: ([a-f0-9]{32})( #})?\n\n', content)
        if not match:
            return False, False, None

        file_chk_expected = match.group(2)
        real_source = content[len(match.group(0)):]

    file_chk_real = source_hash(real_source)

    if file_chk_expected != file_chk_real:
        return True, True, file_chk_real  # checksum failed. Changed by hands?
    else:
        return True, False, file_chk_real


def source_hash(source):
    hash_md5 = hashlib.md5()
    source = source.strip()
    hash_md5.update(source.encode())

    return hash_md5.hexdigest()


def copy_generated_tree(source_path, target_path, glob_expr="**/*"):
    files = []

    with chdir(source_path):
        for file in glob(glob_expr, recursive=True):
            with open(file, 'r') as f:
                files.append((file, f.read()))

    with chdir(target_path):
        for path, source in files:
            write_generated_file(path, source)


def clean_up_generated_files(path, file_list, remove_root=False):
    for file in os.listdir(path):
        if file == 'node_modules':
            continue

        fullpath = os.path.join(path, file)

        if os.path.isdir(fullpath):
            clean_up_generated_files(fullpath, file_list, remove_root=True)
        else:
            if file[0] != '.' and file.split('.')[-1] in ('js', 'jsx', 'py', 'html'):
                generated, changed, file_checksum = is_generated_file(fullpath)

                if generated and (fullpath not in file_list):
                    if changed:
                        print(colored('skip:', 'red', 'on_yellow'), ' ', fullpath)
                    else:
                        print(colored(' - ', 'white', 'on_red'), ' ', fullpath)
                        os.unlink(fullpath)

    if len(os.listdir(path)) == 0 and remove_root:
        os.rmdir(path)
        print(colored('delete:', 'white', 'on_red'), ' ', path)


def extract_files(dst, file_bytes):
    files = zipfile.ZipFile(BytesIO(file_bytes), mode='r', compression=zipfile.ZIP_LZMA)

    for path in files.namelist():
        full_path = os.path.join(dst, path)
        write_generated_file(full_path, files.read(path).decode())

    file_list = [f"./{x}" for x in files.namelist()]
    clean_up_generated_files('.', file_list)


def collect_files(src):
    with chdir(src):
        f = io.BytesIO()
        files = zipfile.ZipFile(f, mode='w', compression=zipfile.ZIP_LZMA)

        for path in get_collect_paths():
            generated, changed, checksum = is_generated_file(path)

            if not generated or changed:
                files.write(path)
        files.close()

        f.seek(0)

    return f


def get_collect_paths():
    paths = []
    paths += list(glob('*.col'))
    # paths += list(glob('react/src/**/*.js', recursive=True))
    # paths += list(glob('react/src/**/*.jsx', recursive=True))
    # paths += list(glob('react/*.js'))
    # paths += list(glob('react/*.json'))

    return set(paths)


def get_watch_paths():
    paths = []
    paths += list(glob('*.col'))
    # paths += list(glob('react/src/**/*.js', recursive=True))
    # paths += list(glob('react/src/**/*.jsx', recursive=True))
    # paths += list(glob('react/*.js'))
    paths += list(glob('react/*.json'))

    return set(paths)


def collect_app_names():
    collections = []
    for filename in os.listdir('.'):
        if os.path.isfile(filename) and filename.endswith('.col'):
            if not re.match('^[a-zA-Z][a-zA-Z0-9_]+\.col$', filename):
                print('Collection file has incorrect name: {}'.format(filename))
            app_name = filename[0:-4]
            collections.append(app_name)
    return collections


def migrate_db(apps, features=None):
    print(colored('> ', 'white', 'on_blue'), 'Migrating database')
    django_command = get_django_command(features)

    for app_name in apps:
        subprocess.run('{} makemigrations {}'.format(django_command, app_name), shell=True, check=True)
        try:
            subprocess.run('{} migrate'.format(django_command), shell=True, check=True)
            # subprocess.run('{} migrate {}'.format(django_command, app_name), shell=True, check=True)
        except subprocess.CalledProcessError:
            pass

        # if cratis
        if django_command == 'django':
            subprocess.run('{} sync_translation_fields_safe'.format(django_command), shell=True, check=True)


def run_django(features=None, run_host='127.0.0.1:8000'):
    print(colored('> ', 'white', 'on_blue'), 'Starting django')
    django_command = get_django_command(features, )
    full_command = '{} runserver {}'.format(django_command, run_host)
    print(full_command)
    return subprocess.Popen(full_command, preexec_fn=os.setsid, shell=True)


def run_webpack():
    print(colored('> ', 'white', 'on_blue'), 'Starting webpack')

    webpack_command = './node_modules/.bin/webpack -w'
    print(webpack_command)
    return subprocess.Popen('{}'.format(webpack_command), shell=True, cwd='react', preexec_fn=os.setsid)


def npm_install():
    if is_req_file_changed('react/package.json'):
        print(colored('> ', 'white', 'on_blue'), 'Installing nodejs dependencies ...')
        cmd = 'npm install'
        print(cmd)
        if subprocess.run('{}'.format(cmd), shell=True, cwd='react', check=True):
            mark_req_file_installed('react/package.json')


def remove_db(apps, features=None):
    django_command = get_django_command(features)

    for app_name in apps:
        filename = '{}.col'.format(app_name)
        if not os.path.exists(filename):
            print('No such collection: {}'.format(app_name))
        app_name = app_name

        print(colored('> ', 'white', 'on_blue'), 'Unapplying migrations')
        subprocess.run('{} migrate {} zero'.format(django_command, app_name), shell=True, check=True)

        print(colored('> ', 'white', 'on_blue'), 'Removing migrations')
        rmtree('{}/migrations'.format(app_name))


def get_django_command(features):
    is_cratis = features and 'cratis' in features
    if is_cratis:
        django_command = 'django'
    else:
        django_command = 'python manage.py'
    return django_command


def is_req_file_changed(filename):
    hash = files_hash([filename])
    cache_filename = f'.zmei-cache/{filename}'
    if os.path.exists(cache_filename):
        with open(cache_filename) as f:
            stored_hash = f.read()
            if stored_hash == hash:
                return False
    return True


def mark_req_file_installed(filename):
    hash = files_hash([filename])
    cache_filename = f'.zmei-cache/{filename}'

    dirname = os.path.dirname(cache_filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(cache_filename, 'w') as f:
        f.write(hash)


def install_deps(django_process):
    if is_req_file_changed('requirements.txt'):
        if django_process:
            os.killpg(os.getpgid(django_process.pid), signal.SIGTERM)
            django_process = None

        print(colored('> ', 'white', 'on_blue'), 'Installing pip dependencies...')
        if subprocess.run('pip install -r requirements.txt', shell=True, check=True):
            mark_req_file_installed('requirements.txt')

    return django_process


def wait_for_file_changes(paths, initial=True, watch=True):
    if initial:
        yield

    if watch:
        initial_hash = files_hash(paths)

        while True:
            sleep(1.0)
            new_hash = files_hash(paths)
            if initial_hash != new_hash:
                initial_hash = new_hash
                yield


def files_hash(paths):
    hash_md5 = hashlib.md5()

    for filename in paths:
        try:
            with open(filename, "rb") as f:
                for chunk in iter(lambda: f.read(2 ** 20), b""):
                    hash_md5.update(chunk)
        except (FileNotFoundError, IOError):
            pass

    return hash_md5.hexdigest()



from setuptools import setup, find_packages

setup(
    name='zmei',
    version='1.1.0',
    packages=find_packages(),

    url='',
    license='Private',
    author='Alex Rudakov',
    author_email='alex@negative.ee',
    description='Genius generator.',
    long_description='',

    install_requires=[
        'click',
        'requests',
        'termcolor',
        'beautifulsoup4',
        'pyparsing',
        'jinja2',
    ],

    entry_points={
        'console_scripts': [
            'zmei-gen = genius_cli.main:run',
        ]
    },
)

